@@ -37,7 +37,8 @@
 // probability of CBox predictions being wrong out of all glyphs loaded in a GlyphSet
 #define CBOX_OVERFLOW_H_PROB 0.2f
 
-RenWindow* window_renderer = NULL;
+static RenWindow **window_list = NULL;
+static size_t window_count = 0;
 static FT_Library library;
 
 // draw_rect_surface is used as a 1x1 surface to simplify ren_draw_rect with blending
@@ -92,6 +93,9 @@ typedef struct RenFont {
   GlyphSet* sets[SUBPIXEL_BITMAPS_CACHED][MAX_LOADABLE_GLYPHSETS];
   float size, space_advance, tab_advance;
   unsigned short baseline, height;
+#ifdef PRAGTICAL_USE_SDL_RENDERER
+  double scale;
+#endif
   ERenFontAntialiasing antialiasing;
   ERenFontHinting hinting;
   unsigned char style;
@@ -109,6 +113,19 @@ typedef struct RenFont {
   char path[];
 } RenFont;
 
+#ifdef PRAGTICAL_USE_SDL_RENDERER
+void update_font_scale(RenWindow *window_renderer, RenFont **fonts) {
+  if (window_renderer == NULL) return;
+  const double surface_scale = renwin_get_surface(window_renderer).scale_x;
+  for (int i = 0; i < FONT_FALLBACK_MAX && fonts[i]; ++i) {
+    if (fonts[i]->scale != surface_scale) {
+      ren_font_group_set_size(fonts, fonts[0]->size, surface_scale);
+      return;
+    }
+  }
+}
+#endif
+
 static const char* utf8_to_codepoint(const char *p, unsigned *dst) {
   const unsigned char *up = (unsigned char*)p;
   unsigned res, n;
@@ -434,7 +451,19 @@ static void font_render_glyph(RenFont* font, unsigned int codepoint) {
     if (!set->surface || metric->x1 > set->surface->w || metric->y1 > set->surface->h) {
       // resize the glyphset with exact pixel sizing
       font_allocate_glyphset(font, glyphset_idx, false);
-      assert(metric->x1 <= set->surface->w && metric->y1 <= set->surface->h);
+      if (!set->surface || metric->x1 > set->surface->w || metric->y1 > set->surface->h) {
+        fprintf(
+          stderr,
+          "Warning: font '%s' at glyphset '%d' not properly resized to "
+          "exact pixel sizing. X1: %d -> W: %d, Y1: %d -> H: %d\n",
+          font->path, glyphset_idx,
+          metric->x1, set->surface ? set->surface->w : 0,
+          metric->y1, set->surface ? set->surface->h : 0
+        );
+      }
+      if (!set->surface) return;
+      // Disable the assert for now since fonts causing this would be rarely used...
+      // assert(metric->x1 <= set->surface->w && metric->y1 <= set->surface->h);
     }
 
     // blit the bitmap onto the surface
@@ -544,7 +573,7 @@ static void font_file_close(FT_Stream stream) {
   }
 }
 
-RenFont* ren_font_load(RenWindow *window_renderer, const char* path, float size, ERenFontAntialiasing antialiasing, ERenFontHinting hinting, unsigned char style) {
+RenFont* ren_font_load(const char* path, float size, ERenFontAntialiasing antialiasing, ERenFontHinting hinting, unsigned char style) {
   RenFont *font = NULL;
   FT_Face face = NULL;
 
@@ -563,13 +592,15 @@ RenFont* ren_font_load(RenWindow *window_renderer, const char* path, float size,
   if (FT_Open_Face(library, &(FT_Open_Args){ .flags = FT_OPEN_STREAM, .stream = &font->stream }, 0, &face))
     goto failure;
 
-  const double surface_scale = renwin_get_surface(window_renderer).scale_x;
-  if (FT_Set_Pixel_Sizes(face, 0, (int)(size*surface_scale)))
+  if (FT_Set_Pixel_Sizes(face, 0, (int)(size)))
     goto failure;
 
   strcpy(font->path, path);
   font->face = face;
   font->size = size;
+#ifdef PRAGTICAL_USE_SDL_RENDERER
+  font->scale = 1.0;
+#endif
   font->height = (short)((face->height / (float)face->units_per_EM) * font->size);
   font->baseline = (short)((face->ascender / (float)face->units_per_EM) * font->size);
   font->antialiasing = antialiasing;
@@ -602,12 +633,12 @@ rwops_failure:
   return NULL;
 }
 
-RenFont* ren_font_copy(RenWindow *window_renderer, RenFont* font, float size, ERenFontAntialiasing antialiasing, ERenFontHinting hinting, int style) {
+RenFont* ren_font_copy(RenFont* font, float size, ERenFontAntialiasing antialiasing, ERenFontHinting hinting, int style) {
   antialiasing = antialiasing == -1 ? font->antialiasing : antialiasing;
   hinting = hinting == -1 ? font->hinting : hinting;
   style = style == -1 ? font->style : style;
 
-  return ren_font_load(window_renderer, font->path, size, antialiasing, hinting, style);
+  return ren_font_load(font->path, size, antialiasing, hinting, style);
 }
 
 const char* ren_font_get_path(RenFont *font) {
@@ -849,13 +880,15 @@ float ren_font_group_get_size(RenFont **fonts) {
   return fonts[0]->size;
 }
 
-void ren_font_group_set_size(RenWindow *window_renderer, RenFont **fonts, float size) {
-  const int surface_scale = renwin_get_surface(window_renderer).scale_x;
+void ren_font_group_set_size(RenFont **fonts, float size, double surface_scale) {
   for (int i = 0; i < FONT_FALLBACK_MAX && fonts[i]; ++i) {
     font_clear_glyph_cache(fonts[i]);
     FT_Face face = fonts[i]->face;
     FT_Set_Pixel_Sizes(face, 0, (int)(size*surface_scale));
     fonts[i]->size = size;
+#ifdef PRAGTICAL_USE_SDL_RENDERER
+    fonts[i]->scale = surface_scale;
+#endif
     fonts[i]->height = (short)((face->height / (float)face->units_per_EM) * size);
     fonts[i]->baseline = (short)((face->ascender / (float)face->units_per_EM) * size);
     FT_Load_Char(face, ' ', font_set_load_options(fonts[i]));
@@ -868,7 +901,7 @@ int ren_font_group_get_height(RenFont **fonts) {
   return fonts[0]->height;
 }
 
-double ren_font_group_get_width(RenWindow *window_renderer, RenFont **fonts, const char *text, size_t len, int *x_offset) {
+double ren_font_group_get_width(RenFont **fonts, const char *text, size_t len, int *x_offset) {
   double width = 0;
   const char* end = text + len;
   bool set_x_offset = x_offset == NULL;
@@ -884,11 +917,14 @@ double ren_font_group_get_width(RenWindow *window_renderer, RenFont **fonts, con
       *x_offset = metric->bitmap_left; // TODO: should this be scaled by the surface scale?
     }
   }
-  const double surface_scale = renwin_get_surface(window_renderer).scale_x;
   if (!set_x_offset) {
     *x_offset = 0;
   }
-  return width / surface_scale;
+#ifdef PRAGTICAL_USE_SDL_RENDERER
+  return width / fonts[0]->scale;
+#else
+  return width;
+#endif
 }
 
 // 4-times unrolled loop
@@ -1108,29 +1144,55 @@ void ren_draw_rect(RenSurface *rs, RenRect rect, RenColor color) {
 }
 
 /*************** Window Management ****************/
-RenWindow* ren_init(SDL_Window *win) {
-  assert(win);
-  int error = FT_Init_FreeType( &library );
-  if ( error ) {
-    fprintf(stderr, "internal font error when starting the application\n");
-    return NULL;
+static void ren_add_window(RenWindow *window_renderer) {
+  window_count += 1;
+  window_list = realloc(window_list, window_count * sizeof(RenWindow*));
+  window_list[window_count-1] = window_renderer;
+}
+
+static void ren_remove_window(RenWindow *window_renderer) {
+  for (size_t i = 0; i < window_count; ++i) {
+    if (window_list[i] == window_renderer) {
+      window_count -= 1;
+      memmove(&window_list[i], &window_list[i+1], window_count - i);
+      return;
+    }
   }
+}
+
+int ren_init(void) {
+  int err;
+
+  draw_rect_surface = SDL_CreateRGBSurface(0, 1, 1, 32,
+                       0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF);
+
+  if ((err = FT_Init_FreeType( &library )))
+    return err;
+
+  return 0;
+}
+
+void ren_free(void) {
+   SDL_FreeSurface(draw_rect_surface);
+}
+
+RenWindow* ren_create(SDL_Window *win) {
+  assert(win);
   RenWindow* window_renderer = calloc(1, sizeof(RenWindow));
 
   window_renderer->window = win;
   renwin_init_surface(window_renderer);
   renwin_init_command_buf(window_renderer);
   renwin_clip_to_surface(window_renderer);
-  draw_rect_surface = SDL_CreateRGBSurface(0, 1, 1, 32,
-                       0xFF000000, 0x00FF0000, 0x0000FF00, 0x000000FF);
 
+  ren_add_window(window_renderer);
   return window_renderer;
 }
 
-void ren_free(RenWindow* window_renderer) {
+void ren_destroy(RenWindow* window_renderer) {
   assert(window_renderer);
+  ren_remove_window(window_renderer);
   renwin_free(window_renderer);
-  SDL_FreeSurface(draw_rect_surface);
   free(window_renderer->command_buf);
   window_renderer->command_buf = NULL;
   window_renderer->command_buf_size = 0;
@@ -1141,14 +1203,14 @@ void ren_resize_window(RenWindow *window_renderer) {
   renwin_resize_surface(window_renderer);
 }
 
-
+// TODO: Does not work nicely with multiple windows
 void ren_update_rects(RenWindow *window_renderer, RenRect *rects, int count) {
   static bool initial_frame = true;
+  renwin_update_rects(window_renderer, rects, count);
   if (initial_frame) {
     renwin_show_window(window_renderer);
     initial_frame = false;
   }
-  renwin_update_rects(window_renderer, rects, count);
 }
 
 
@@ -1159,8 +1221,12 @@ void ren_set_clip_rect(RenWindow *window_renderer, RenRect rect) {
 
 void ren_get_size(RenWindow *window_renderer, int *x, int *y) {
   RenSurface rs = renwin_get_surface(window_renderer);
-  *x = rs.surface->w / rs.scale_x;
-  *y = rs.surface->h / rs.scale_y;
+  *x = rs.surface->w;
+  *y = rs.surface->h;
+#ifdef PRAGTICAL_USE_SDL_RENDERER
+  *x /= rs.scale_x;
+  *y /= rs.scale_y;
+#endif
 }
 
 
@@ -1172,3 +1238,24 @@ float ren_get_scale_factor(SDL_Window *win) {
   float scale = (float) w_pixels / (float) w_points;
   return roundf(scale * 100) / 100;
 }
+
+size_t ren_get_window_list(RenWindow ***window_list_dest) {
+  *window_list_dest = window_list;
+  return window_count;
+}
+
+RenWindow* ren_find_window(SDL_Window *window) {
+  for (size_t i = 0; i < window_count; ++i) {
+    RenWindow* window_renderer = window_list[i];
+    if (window_renderer->window == window) {
+      return window_renderer;
+    }
+  }
+
+  return NULL;
+}
+
+RenWindow* ren_find_window_from_id(uint32_t id) {
+  SDL_Window *window = SDL_GetWindowFromID(id);
+  return ren_find_window(window);
+}
